@@ -1,21 +1,38 @@
 # -*- coding: utf-8 -*-
+#
+# Copyright (C) 2019 Radim Rehurek <me@radimrehurek.com>
+#
+# This code is distributed under the terms and conditions
+# from the MIT License (MIT).
+#
 """Implements file-like objects for reading and writing from/to S3."""
 
 import io
 import contextlib
 import functools
 import logging
+import time
 import warnings
 
 import boto3
 import botocore.client
+import botocore.exceptions
 import six
-import sys
 
 import smart_open.bytebuffer
 
 logger = logging.getLogger(__name__)
 
+# AWS Lambda environments do not support multiprocessing.Queue or multiprocessing.Pool.
+# However they do support Threads and therefore concurrent.futures's ThreadPoolExecutor.
+# We use this flag to allow python 2 backward compatibility, where concurrent.futures doesn't exist.
+_CONCURRENT_FUTURES = False
+try:
+    import concurrent.futures
+    _CONCURRENT_FUTURES = True
+except ImportError:
+    warnings.warn("concurrent.futures could not be imported and won't be used")
+
 # Multiprocessing is unavailable in App Engine (and possibly other sandboxes).
 # The only method currently relying on it is iter_bucket, which is instructed
 # whether to use it by the MULTIPROCESSING flag.
@@ -36,11 +53,6 @@ WRITE_BINARY = 'wb'
 MODES = (READ_BINARY, WRITE_BINARY)
 """Allowed I/O modes for working with S3."""
 
-_BINARY_TYPES = (six.binary_type, bytearray)
-"""Allowed binary buffer types for writing to the underlying S3 stream"""
-if sys.version_info >= (2, 7):
-    _BINARY_TYPES = (six.binary_type, bytearray, memoryview)
-
 BINARY_NEWLINE = b'\n'
 
 SUPPORTED_SCHEMES = ("s3", "s3n", 's3u', "s3a")
@@ -52,6 +64,9 @@ CURRENT = 1
 END = 2
 WHENCE_CHOICES = [START, CURRENT, END]
 
+_UPLOAD_ATTEMPTS = 6
+_SLEEP_SECONDS = 10
+
 
 def clamp(value, minval, maxval):
     return max(min(value, maxval), minval)
@@ -70,11 +85,15 @@ def open(
         bucket_id,
         key_id,
         mode,
+        version_id=None,
         buffer_size=DEFAULT_BUFFER_SIZE,
         min_part_size=DEFAULT_MIN_PART_SIZE,
         session=None,
         resource_kwargs=None,
         multipart_upload_kwargs=None,
+        multipart_upload=True,
+        singlepart_upload_kwargs=None,
+        object_kwargs=None,
         ):
     """Open an S3 object for reading or writing.
 
@@ -97,111 +116,187 @@ def open(
     multipart_upload_kwargs: dict, optional
         Additional parameters to pass to boto3's initiate_multipart_upload function.
         For writing only.
+    singlepart_upload_kwargs: dict, optional
+        Additional parameters to pass to boto3's S3.Object.put function when using single
+        part upload.
+        For writing only.
+    multipart_upload: bool, optional
+        Default: `True`
+        If set to `True`, will use multipart upload for writing to S3. If set
+        to `False`, S3 upload will use the S3 Single-Part Upload API, which
+        is more ideal for small file sizes.
+        For writing only.
+    version_id: str, optional
+        Version of the object, used when reading object.
+        If None, will fetch the most recent version.
+    object_kwargs: dict, optional
+        Additional parameters to pass to boto3's object.get function.
+        Used during reading only.
 
     """
     logger.debug('%r', locals())
     if mode not in MODES:
         raise NotImplementedError('bad mode: %r expected one of %r' % (mode, MODES))
 
-    if resource_kwargs is None:
-        resource_kwargs = {}
-    if multipart_upload_kwargs is None:
-        multipart_upload_kwargs = {}
+    if (mode == WRITE_BINARY) and (version_id is not None):
+        raise ValueError("version_id must be None when writing")
 
     if mode == READ_BINARY:
-        fileobj = SeekableBufferedInputBase(
+        fileobj = Reader(
             bucket_id,
             key_id,
+            version_id=version_id,
             buffer_size=buffer_size,
             session=session,
             resource_kwargs=resource_kwargs,
+            object_kwargs=object_kwargs,
         )
     elif mode == WRITE_BINARY:
-        fileobj = BufferedOutputBase(
-            bucket_id,
-            key_id,
-            min_part_size=min_part_size,
-            session=session,
-            multipart_upload_kwargs=multipart_upload_kwargs,
-            resource_kwargs=resource_kwargs,
-        )
+        if multipart_upload:
+            fileobj = MultipartWriter(
+                bucket_id,
+                key_id,
+                min_part_size=min_part_size,
+                session=session,
+                upload_kwargs=multipart_upload_kwargs,
+                resource_kwargs=resource_kwargs,
+            )
+        else:
+            fileobj = SinglepartWriter(
+                bucket_id,
+                key_id,
+                session=session,
+                upload_kwargs=singlepart_upload_kwargs,
+                resource_kwargs=resource_kwargs,
+            )
     else:
         assert False, 'unexpected mode: %r' % mode
-
     return fileobj
 
 
-class RawReader(object):
-    """Read an S3 object."""
-    def __init__(self, s3_object):
-        self.position = 0
-        self._object = s3_object
-        self._body = s3_object.get()['Body']
-
-    def read(self, size=-1):
-        if size == -1:
-            return self._body.read()
-        return self._body.read(size)
+def _get(s3_object, version=None, **kwargs):
+    if version is not None:
+        kwargs['VersionId'] = version
+    try:
+        return s3_object.get(**kwargs)
+    except botocore.client.ClientError as error:
+        raise IOError(
+            'unable to access bucket: %r key: %r version: %r error: %s' % (
+                s3_object.bucket_name, s3_object.key, version, error
+            )
+        )
 
 
-class SeekableRawReader(object):
-    """Read an S3 object."""
+class _SeekableRawReader(object):
+    """Read an S3 object.
 
-    def __init__(self, s3_object, content_length):
+    This class is internal to the S3 submodule.
+    """
+
+    def __init__(self, s3_object, content_length, version_id=None, object_kwargs=None):
         self._object = s3_object
         self._content_length = content_length
-        self.seek(0)
+        self._version_id = version_id
+        self._position = 0
+        self._body = None
+        self._object_kwargs = object_kwargs if object_kwargs else {}
 
     def seek(self, position):
         """Seek to the specified position (byte offset) in the S3 key.
 
         :param int position: The byte offset from the beginning of the key.
         """
+        #
+        # Close old body explicitly.
+        # When first seek() after __init__(), self._body is not exist.
+        #
+        if self._body is not None:
+            self._body.close()
+        self._body = None
         self._position = position
+
+    def _load_body(self):
+        """Build a continuous connection with the remote peer starts from the current postion.
+        """
         range_string = make_range_string(self._position)
         logger.debug('content_length: %r range_string: %r', self._content_length, range_string)
 
-        #
-        # Close old body explicitly.
-        # When first seek(), self._body is not exist. Catch the exception and do nothing.
-        #
-        try:
-            self._body.close()
-        except AttributeError:
-            pass
-
-        if position == self._content_length == 0 or position == self._content_length:
+        if self._position == self._content_length == 0 or self._position == self._content_length:
             #
             # When reading, we can't seek to the first byte of an empty file.
             # Similarly, we can't seek past the last byte.  Do nothing here.
             #
             self._body = io.BytesIO()
         else:
-            self._body = self._object.get(Range=range_string)['Body']
+            self._body = _get(
+                self._object,
+                version=self._version_id,
+                Range=range_string,
+                **self._object_kwargs
+            )['Body']
 
-    def read(self, size=-1):
-        if self._position >= self._content_length:
-            return b''
+    def _read_from_body(self, size=-1):
         if size == -1:
             binary = self._body.read()
         else:
             binary = self._body.read(size)
+        return binary
+
+    def read(self, size=-1):
+        """Read from the continuous connection with the remote peer."""
+        if self._position >= self._content_length:
+            return b''
+        if self._body is None:
+            # When the first read() after __init__() or seek(), self._body is not exist.
+            self._load_body()
+
+        try:
+            binary = self._read_from_body(size)
+        except botocore.exceptions.IncompleteReadError:
+            # The underlying connection of the self._body was closed by the remote peer.
+            self._load_body()
+            binary = self._read_from_body(size)
         self._position += len(binary)
         return binary
 
 
-class BufferedInputBase(io.BufferedIOBase):
-    def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=None):
+class Reader(io.BufferedIOBase):
+    """Reads bytes from S3.
+
+    Implements the io.BufferedIOBase interface of the standard library."""
+
+    def __init__(self, bucket, key, version_id=None, buffer_size=DEFAULT_BUFFER_SIZE,
+                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=None,
+                 object_kwargs=None):
+
+        self._buffer_size = buffer_size
+
         if session is None:
             session = boto3.Session()
         if resource_kwargs is None:
             resource_kwargs = {}
+        if object_kwargs is None:
+            object_kwargs = {}
+
+        self._session = session
+        self._resource_kwargs = resource_kwargs
+        self._object_kwargs = object_kwargs
 
         s3 = session.resource('s3', **resource_kwargs)
         self._object = s3.Object(bucket, key)
-        self._raw_reader = RawReader(self._object)
-        self._content_length = self._object.content_length
+        self._version_id = version_id
+        self._content_length = _get(
+            self._object,
+            version=self._version_id,
+            **self._object_kwargs
+        )['ContentLength']
+
+        self._raw_reader = _SeekableRawReader(
+            self._object,
+            self._content_length,
+            self._version_id,
+            self._object_kwargs,
+        )
         self._current_pos = 0
         self._buffer = smart_open.bytebuffer.ByteBuffer(buffer_size)
         self._eof = False
@@ -213,8 +308,9 @@ class BufferedInputBase(io.BufferedIOBase):
         self.raw = None
 
     #
-    # Override some methods from io.IOBase.
+    # io.BufferedIOBase methods.
     #
+
     def close(self):
         """Flush and close this stream."""
         logger.debug("close: called")
@@ -224,16 +320,6 @@ class BufferedInputBase(io.BufferedIOBase):
         """Return True if the stream can be read from."""
         return True
 
-    def seekable(self):
-        return False
-
-    #
-    # io.BufferedIOBase methods.
-    #
-    def detach(self):
-        """Unsupported."""
-        raise io.UnsupportedOperation
-
     def read(self, size=-1):
         """Read up to size bytes from the object and return them."""
         if size == 0:
@@ -255,10 +341,6 @@ class BufferedInputBase(io.BufferedIOBase):
         if self._eof:
             return self._read_from_buffer()
 
-        #
-        # Fill our buffer to the required size.
-        #
-        # logger.debug('filling %r byte-long buffer up to %r bytes', len(self._buffer), size)
         self._fill_buffer(size)
         return self._read_from_buffer(size)
 
@@ -279,82 +361,22 @@ class BufferedInputBase(io.BufferedIOBase):
         """Read up to and including the next newline.  Returns the bytes read."""
         if limit != -1:
             raise NotImplementedError('limits other than -1 not implemented yet')
-        the_line = io.BytesIO()
+
+        #
+        # A single line may span multiple buffers.
+        #
+        line = io.BytesIO()
         while not (self._eof and len(self._buffer) == 0):
-            #
-            # In the worst case, we're reading the unread part of self._buffer
-            # twice here, once in the if condition and once when calling index.
-            #
-            # This is sub-optimal, but better than the alternative: wrapping
-            # .index in a try..except, because that is slower.
-            #
-            remaining_buffer = self._buffer.peek()
-            if self._line_terminator in remaining_buffer:
-                next_newline = remaining_buffer.index(self._line_terminator)
-                the_line.write(self._read_from_buffer(next_newline + 1))
+            line_part = self._buffer.readline(self._line_terminator)
+            line.write(line_part)
+            self._current_pos += len(line_part)
+
+            if line_part.endswith(self._line_terminator):
                 break
             else:
-                the_line.write(self._read_from_buffer())
                 self._fill_buffer()
-        return the_line.getvalue()
 
-    def terminate(self):
-        """Do nothing."""
-        pass
-
-    #
-    # Internal methods.
-    #
-    def _read_from_buffer(self, size=-1):
-        """Remove at most size bytes from our buffer and return them."""
-        # logger.debug('reading %r bytes from %r byte-long buffer', size, len(self._buffer))
-        size = size if size >= 0 else len(self._buffer)
-        part = self._buffer.read(size)
-        self._current_pos += len(part)
-        # logger.debug('part: %r', part)
-        return part
-
-    def _fill_buffer(self, size=-1):
-        size = size if size >= 0 else self._buffer._chunk_size
-        while len(self._buffer) < size and not self._eof:
-            bytes_read = self._buffer.fill(self._raw_reader)
-            if bytes_read == 0:
-                logger.debug('reached EOF while filling buffer')
-                self._eof = True
-
-
-class SeekableBufferedInputBase(BufferedInputBase):
-    """Reads bytes from S3.
-
-    Implements the io.BufferedIOBase interface of the standard library."""
-
-    def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=None):
-        if session is None:
-            session = boto3.Session()
-        if resource_kwargs is None:
-            resource_kwargs = {}
-        s3 = session.resource('s3', **resource_kwargs)
-        self._object = s3.Object(bucket, key)
-
-        try:
-            self._content_length = self._object.content_length
-        except botocore.client.ClientError:
-            raise ValueError(
-                '%r does not exist in the bucket %r, '
-                'or is forbidden for access' % (key, bucket)
-            )
-
-        self._raw_reader = SeekableRawReader(self._object, self._content_length)
-        self._current_pos = 0
-        self._buffer = smart_open.bytebuffer.ByteBuffer(buffer_size)
-        self._eof = False
-        self._line_terminator = line_terminator
-
-        #
-        # This member is part of the io.BufferedIOBase interface.
-        #
-        self.raw = None
+        return line.getvalue()
 
     def seekable(self):
         """If False, seek(), tell() and truncate() will raise IOError.
@@ -396,9 +418,74 @@ class SeekableBufferedInputBase(BufferedInputBase):
         """Unsupported."""
         raise io.UnsupportedOperation
 
+    def detach(self):
+        """Unsupported."""
+        raise io.UnsupportedOperation
 
-class BufferedOutputBase(io.BufferedIOBase):
-    """Writes bytes to S3.
+    def terminate(self):
+        """Do nothing."""
+        pass
+
+    def to_boto3(self):
+        """Create an **independent** `boto3.s3.Object` instance that points to
+        the same resource as this instance.
+
+        The created instance will re-use the session and resource parameters of
+        the current instance, but it will be independent: changes to the
+        `boto3.s3.Object` may not necessary affect the current instance.
+
+        """
+        s3 = self._session.resource('s3', **self._resource_kwargs)
+        return s3.Object(self._object.bucket_name, self._object.key)
+
+    #
+    # Internal methods.
+    #
+    def _read_from_buffer(self, size=-1):
+        """Remove at most size bytes from our buffer and return them."""
+        # logger.debug('reading %r bytes from %r byte-long buffer', size, len(self._buffer))
+        size = size if size >= 0 else len(self._buffer)
+        part = self._buffer.read(size)
+        self._current_pos += len(part)
+        # logger.debug('part: %r', part)
+        return part
+
+    def _fill_buffer(self, size=-1):
+        size = max(size, self._buffer._chunk_size)
+        while len(self._buffer) < size and not self._eof:
+            bytes_read = self._buffer.fill(self._raw_reader)
+            if bytes_read == 0:
+                logger.debug('reached EOF while filling buffer')
+                self._eof = True
+
+    def __str__(self):
+        return "smart_open.s3.Reader(%r, %r)" % (
+            self._object.bucket_name, self._object.key
+        )
+
+    def __repr__(self):
+        return (
+            "smart_open.s3.Reader("
+            "bucket=%r, "
+            "key=%r, "
+            "version_id=%r, "
+            "buffer_size=%r, "
+            "line_terminator=%r, "
+            "session=%r, "
+            "resource_kwargs=%r)"
+        ) % (
+            self._object.bucket_name,
+            self._object.key,
+            self._version_id,
+            self._buffer_size,
+            self._line_terminator,
+            self._session,
+            self._resource_kwargs,
+        )
+
+
+class MultipartWriter(io.BufferedIOBase):
+    """Writes bytes to S3 using the multi part API.
 
     Implements the io.BufferedIOBase interface of the standard library."""
 
@@ -409,7 +496,7 @@ class BufferedOutputBase(io.BufferedIOBase):
             min_part_size=DEFAULT_MIN_PART_SIZE,
             session=None,
             resource_kwargs=None,
-            multipart_upload_kwargs=None,
+            upload_kwargs=None,
             ):
         if min_part_size < MIN_MIN_PART_SIZE:
             logger.warning("S3 requires minimum part size >= 5MB; \
@@ -419,16 +506,25 @@ multipart upload may fail")
             session = boto3.Session()
         if resource_kwargs is None:
             resource_kwargs = {}
-        if multipart_upload_kwargs is None:
-            multipart_upload_kwargs = {}
+        if upload_kwargs is None:
+            upload_kwargs = {}
+
+        self._session = session
+        self._resource_kwargs = resource_kwargs
+        self._upload_kwargs = upload_kwargs
 
         s3 = session.resource('s3', **resource_kwargs)
         try:
             self._object = s3.Object(bucket, key)
             self._min_part_size = min_part_size
-            self._mp = self._object.initiate_multipart_upload(**multipart_upload_kwargs)
-        except botocore.client.ClientError:
-            raise ValueError('the bucket %r does not exist, or is forbidden for access' % bucket)
+            partial = functools.partial(self._object.initiate_multipart_upload, **self._upload_kwargs)
+            self._mp = _retry_if_failed(partial)
+        except botocore.client.ClientError as error:
+            raise ValueError(
+                'the bucket %r does not exist, or is forbidden for access (%r)' % (
+                    bucket, error
+                )
+            )
 
         self._buf = io.BytesIO()
         self._total_bytes = 0
@@ -452,7 +548,8 @@ multipart upload may fail")
             self._upload_next_part()
 
         if self._total_bytes and self._mp:
-            self._mp.complete(MultipartUpload={'Parts': self._parts})
+            partial = functools.partial(self._mp.complete, MultipartUpload={'Parts': self._parts})
+            _retry_if_failed(partial)
             logger.debug("completed multipart upload")
         elif self._mp:
             #
@@ -465,7 +562,6 @@ multipart upload may fail")
             logger.info("empty input, ignoring multipart upload")
             assert self._mp, "no multipart upload in progress"
             self._mp.abort()
-
             self._object.put(Body=b'')
         self._mp = None
         logger.debug("successfully closed")
@@ -489,22 +585,21 @@ multipart upload may fail")
         raise io.UnsupportedOperation("detach() not supported")
 
     def write(self, b):
-        """Write the given bytes (binary string) to the S3 file.
+        """Write the given buffer (bytes, bytearray, memoryview or any buffer
+        interface implementation) to the S3 file.
+
+        For more information about buffers, see https://docs.python.org/3/c-api/buffer.html
 
         There's buffering happening under the covers, so this may not actually
         do any HTTP transfer right away."""
 
-        if not isinstance(b, _BINARY_TYPES):
-            raise TypeError(
-                "input must be one of %r, got: %r" % (_BINARY_TYPES, type(b)))
-
-        self._buf.write(b)
-        self._total_bytes += len(b)
+        length = self._buf.write(b)
+        self._total_bytes += length
 
         if self._buf.tell() >= self._min_part_size:
             self._upload_next_part()
 
-        return len(b)
+        return length
 
     def terminate(self):
         """Cancel the underlying multipart upload."""
@@ -512,6 +607,18 @@ multipart upload may fail")
         self._mp.abort()
         self._mp = None
 
+    def to_boto3(self):
+        """Create an **independent** `boto3.s3.Object` instance that points to
+        the same resource as this instance.
+
+        The created instance will re-use the session and resource parameters of
+        the current instance, but it will be independent: changes to the
+        `boto3.s3.Object` may not necessary affect the current instance.
+
+        """
+        s3 = self._session.resource('s3', **self._resource_kwargs)
+        return s3.Object(self._object.bucket_name, self._object.key)
+
     #
     # Internal methods.
     #
@@ -521,7 +628,15 @@ multipart upload may fail")
                     part_num, self._buf.tell(), self._total_bytes / 1024.0 ** 3)
         self._buf.seek(0)
         part = self._mp.Part(part_num)
-        upload = part.upload(Body=self._buf)
+
+        #
+        # Network problems in the middle of an upload are particularly
+        # troublesome.  We don't want to abort the entire upload just because
+        # of a temporary connection problem, so this part needs to be
+        # especially robust.
+        #
+        upload = _retry_if_failed(functools.partial(part.upload, Body=self._buf))
+
         self._parts.append({'ETag': upload['ETag'], 'PartNumber': part_num})
         logger.debug("upload of part #%i finished" % part_num)
 
@@ -537,9 +652,190 @@ multipart upload may fail")
         else:
             self.close()
 
+    def __str__(self):
+        return "smart_open.s3.MultipartWriter(%r, %r)" % (
+            self._object.bucket_name, self._object.key,
+        )
 
-def iter_bucket(bucket_name, prefix='', accept_key=None,
-                key_limit=None, workers=16, retries=3):
+    def __repr__(self):
+        return (
+            "smart_open.s3.MultipartWriter(bucket=%r, key=%r, "
+            "min_part_size=%r, session=%r, resource_kwargs=%r, upload_kwargs=%r)"
+        ) % (
+            self._object.bucket_name,
+            self._object.key,
+            self._min_part_size,
+            self._session,
+            self._resource_kwargs,
+            self._upload_kwargs,
+        )
+
+
+class SinglepartWriter(io.BufferedIOBase):
+    """Writes bytes to S3 using the single part API.
+
+    Implements the io.BufferedIOBase interface of the standard library.
+
+    This class buffers all of its input in memory until its `close` method is called. Only then will
+    the data be written to S3 and the buffer is released."""
+
+    def __init__(
+            self,
+            bucket,
+            key,
+            session=None,
+            resource_kwargs=None,
+            upload_kwargs=None,
+            ):
+
+        self._session = session
+        self._resource_kwargs = resource_kwargs
+
+        if session is None:
+            session = boto3.Session()
+        if resource_kwargs is None:
+            resource_kwargs = {}
+        if upload_kwargs is None:
+            upload_kwargs = {}
+
+        self._upload_kwargs = upload_kwargs
+
+        s3 = session.resource('s3', **resource_kwargs)
+        try:
+            self._object = s3.Object(bucket, key)
+            s3.meta.client.head_bucket(Bucket=bucket)
+        except botocore.client.ClientError:
+            raise ValueError('the bucket %r does not exist, or is forbidden for access' % bucket)
+
+        self._buf = io.BytesIO()
+        self._total_bytes = 0
+
+        #
+        # This member is part of the io.BufferedIOBase interface.
+        #
+        self.raw = None
+
+    def flush(self):
+        pass
+
+    #
+    # Override some methods from io.IOBase.
+    #
+    def close(self):
+        if self._buf is None:
+            return
+
+        self._buf.seek(0)
+
+        try:
+            self._object.put(Body=self._buf, **self._upload_kwargs)
+        except botocore.client.ClientError:
+            raise ValueError(
+                'the bucket %r does not exist, or is forbidden for access' % self._object.bucket_name)
+
+        logger.debug("direct upload finished")
+        self._buf = None
+
+    @property
+    def closed(self):
+        return self._buf is None
+
+    def writable(self):
+        """Return True if the stream supports writing."""
+        return True
+
+    def tell(self):
+        """Return the current stream position."""
+        return self._total_bytes
+
+    #
+    # io.BufferedIOBase methods.
+    #
+    def detach(self):
+        raise io.UnsupportedOperation("detach() not supported")
+
+    def write(self, b):
+        """Write the given buffer (bytes, bytearray, memoryview or any buffer
+        interface implementation) into the buffer. Content of the buffer will be
+        written to S3 on close as a single-part upload.
+
+        For more information about buffers, see https://docs.python.org/3/c-api/buffer.html"""
+
+        length = self._buf.write(b)
+        self._total_bytes += length
+        return length
+
+    def terminate(self):
+        """Nothing to cancel in single-part uploads."""
+        return
+
+    #
+    # Internal methods.
+    #
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        if exc_type is not None:
+            self.terminate()
+        else:
+            self.close()
+
+    def __str__(self):
+        return "smart_open.s3.SinglepartWriter(%r, %r)" % (self._object.bucket_name, self._object.key)
+
+    def __repr__(self):
+        return (
+            "smart_open.s3.SinglepartWriter(bucket=%r, key=%r, session=%r, "
+            "resource_kwargs=%r, upload_kwargs=%r)"
+        ) % (
+            self._object.bucket_name,
+            self._object.key,
+            self._session,
+            self._resource_kwargs,
+            self._upload_kwargs,
+        )
+
+
+def _retry_if_failed(
+        partial,
+        attempts=_UPLOAD_ATTEMPTS,
+        sleep_seconds=_SLEEP_SECONDS,
+        exceptions=(botocore.exceptions.EndpointConnectionError, )):
+    for attempt in range(attempts):
+        try:
+            return partial()
+        except exceptions:
+            logger.critical(
+                'Unable to connect to the endpoint. Check your network connection. '
+                'Sleeping and retrying %d more times '
+                'before giving up.' % (attempts - attempt - 1)
+            )
+            time.sleep(sleep_seconds)
+    else:
+        logger.critical('Unable to connect to the endpoint. Giving up.')
+        raise IOError('Unable to connect to the endpoint after %d attempts' % attempts)
+
+
+#
+# For backward compatibility
+#
+SeekableBufferedInputBase = Reader
+BufferedOutputBase = MultipartWriter
+
+
+def _accept_all(key):
+    return True
+
+
+def iter_bucket(
+        bucket_name,
+        prefix='',
+        accept_key=None,
+        key_limit=None,
+        workers=16,
+        retries=3,
+        **session_kwargs):
     """
     Iterate and download all S3 objects under `s3://bucket_name/prefix`.
 
@@ -559,6 +855,11 @@ def iter_bucket(bucket_name, prefix='', accept_key=None,
         The number of subprocesses to use.
     retries: int, optional
         The number of time to retry a failed download.
+    session_kwargs: dict, optional
+        Keyword arguments to pass when creating a new session.
+        For a list of available names and values, see:
+        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session
+
 
     Yields
     ------
@@ -577,7 +878,9 @@ def iter_bucket(bucket_name, prefix='', accept_key=None,
     --------
 
       >>> # get all JSON files under "mybucket/foo/"
-      >>> for key, content in iter_bucket(bucket_name, prefix='foo/', accept_key=lambda key: key.endswith('.json')):
+      >>> for key, content in iter_bucket(
+      ...         bucket_name, prefix='foo/',
+      ...         accept_key=lambda key: key.endswith('.json')):
       ...     print key, len(content)
 
       >>> # limit to 10k files, using 32 parallel workers (default is 16)
@@ -585,7 +888,7 @@ def iter_bucket(bucket_name, prefix='', accept_key=None,
       ...     print key, len(content)
     """
     if accept_key is None:
-        accept_key = lambda key: True
+        accept_key = _accept_all
 
     #
     # If people insist on giving us bucket instances, silently extract the name
@@ -597,8 +900,16 @@ def iter_bucket(bucket_name, prefix='', accept_key=None,
         pass
 
     total_size, key_no = 0, -1
-    key_iterator = _list_bucket(bucket_name, prefix=prefix, accept_key=accept_key)
-    download_key = functools.partial(_download_key, bucket_name=bucket_name, retries=retries)
+    key_iterator = _list_bucket(
+        bucket_name,
+        prefix=prefix,
+        accept_key=accept_key,
+        **session_kwargs)
+    download_key = functools.partial(
+        _download_key,
+        bucket_name=bucket_name,
+        retries=retries,
+        **session_kwargs)
 
     with _create_process_pool(processes=workers) as pool:
         result_iterator = pool.imap_unordered(download_key, key_iterator)
@@ -617,17 +928,23 @@ def iter_bucket(bucket_name, prefix='', accept_key=None,
     logger.info("processed %i keys, total size %i" % (key_no + 1, total_size))
 
 
-def _list_bucket(bucket_name, prefix='', accept_key=lambda k: True):
-    client = boto3.client('s3')
+def _list_bucket(
+        bucket_name,
+        prefix='',
+        accept_key=lambda k: True,
+        **session_kwargs):
+    session = boto3.session.Session(**session_kwargs)
+    client = session.client('s3')
     ctoken = None
 
     while True:
         # list_objects_v2 doesn't like a None value for ContinuationToken
         # so we don't set it if we don't have one.
         if ctoken:
-            response = client.list_objects_v2(Bucket=bucket_name, Prefix=prefix, ContinuationToken=ctoken)
+            kwargs = dict(Bucket=bucket_name, Prefix=prefix, ContinuationToken=ctoken)
         else:
-            response = client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
+            kwargs = dict(Bucket=bucket_name, Prefix=prefix)
+        response = client.list_objects_v2(**kwargs)
         try:
             content = response['Contents']
         except KeyError:
@@ -642,18 +959,19 @@ def _list_bucket(bucket_name, prefix='', accept_key=lambda k: True):
             break
 
 
-def _download_key(key_name, bucket_name=None, retries=3):
+def _download_key(key_name, bucket_name=None, retries=3, **session_kwargs):
     if bucket_name is None:
         raise ValueError('bucket_name may not be None')
 
     #
     # https://geekpete.com/blog/multithreading-boto3/
     #
-    session = boto3.session.Session()
+    session = boto3.session.Session(**session_kwargs)
     s3 = session.resource('s3')
     bucket = s3.Bucket(bucket_name)
 
-    # Sometimes, https://github.com/boto/boto/issues/2409 can happen because of network issues on either side.
+    # Sometimes, https://github.com/boto/boto/issues/2409 can happen
+    # because of network issues on either side.
     # Retry up to 3 times to ensure its not a transient issue.
     for x in range(retries + 1):
         try:
@@ -687,11 +1005,28 @@ class DummyPool(object):
         pass
 
 
+class ConcurrentFuturesPool(object):
+    """A class that mimics multiprocessing.pool.Pool but uses concurrent futures instead of processes."""
+    def __init__(self, max_workers):
+        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers)
+
+    def imap_unordered(self, function, items):
+        futures = [self.executor.submit(function, item) for item in items]
+        for future in concurrent.futures.as_completed(futures):
+            yield future.result()
+
+    def terminate(self):
+        self.executor.shutdown(wait=True)
+
+
 @contextlib.contextmanager
 def _create_process_pool(processes=1):
     if _MULTIPROCESSING and processes:
-        logger.info("creating pool with %i workers", processes)
+        logger.info("creating multiprocessing pool with %i workers", processes)
         pool = multiprocessing.pool.Pool(processes=processes)
+    elif _CONCURRENT_FUTURES and processes:
+        logger.info("creating concurrent futures pool with %i workers", processes)
+        pool = ConcurrentFuturesPool(max_workers=processes)
     else:
         logger.info("creating dummy pool")
         pool = DummyPool()
