@@ -26,11 +26,9 @@ import os.path as P
 import warnings
 import sys
 
-import boto
 import boto3
 import six
 
-from boto.compat import urlsplit
 from six.moves.urllib import parse as urlparse
 
 #
@@ -42,6 +40,7 @@ import smart_open.hdfs as smart_open_hdfs
 import smart_open.webhdfs as smart_open_webhdfs
 import smart_open.http as smart_open_http
 import smart_open.ssh as smart_open_ssh
+import smart_open.gcs as smart_open_gcs
 
 from smart_open import doctools
 
@@ -123,18 +122,20 @@ Uri = collections.namedtuple(
         'uri_path',
         'bucket_id',
         'key_id',
+        'blob_id',
         'port',
         'host',
         'ordinary_calling_format',
         'access_id',
         'access_secret',
         'user',
+        'password',
     )
 )
 """Represents all the options that we parse from user input.
 
 Some of the above options only make sense for certain protocols, e.g.
-bucket_id is only for S3.
+bucket_id is only for S3 and GCS.
 """
 #
 # Set the default values for all Uri fields to be None.  This allows us to only
@@ -363,7 +364,10 @@ def open(
     return decoded
 
 
-open.__doc__ = open.__doc__ % {
+#
+# The docstring can be None if -OO was passed to the interpreter.
+#
+open.__doc__ = None if open.__doc__ is None else open.__doc__ % {
     's3': doctools.to_docstring(
         doctools.extract_kwargs(smart_open_s3.open.__doc__),
         lpad=u'    ',
@@ -380,6 +384,10 @@ open.__doc__ = open.__doc__ % {
         doctools.extract_kwargs(smart_open_ssh.open.__doc__),
         lpad=u'    ',
     ),
+    'gcs': doctools.to_docstring(
+        doctools.extract_kwargs(smart_open_gcs.open.__doc__),
+        lpad=u'    ',
+    ),
     'examples': doctools.extract_examples_from_readme_rst(),
 }
 
@@ -554,6 +562,8 @@ def _open_binary_stream(uri, mode, transport_params):
                 host=parsed_uri.host,
                 user=parsed_uri.user,
                 port=parsed_uri.port,
+                password=parsed_uri.password,
+                transport_params=transport_params,
             )
             return fobj, filename
         elif parsed_uri.scheme in smart_open_s3.SUPPORTED_SCHEMES:
@@ -563,7 +573,8 @@ def _open_binary_stream(uri, mode, transport_params):
             return smart_open_hdfs.open(parsed_uri.uri_path, mode), filename
         elif parsed_uri.scheme == "webhdfs":
             kw = _check_kwargs(smart_open_webhdfs.open, transport_params)
-            return smart_open_webhdfs.open(parsed_uri.uri_path, mode, **kw), filename
+            http_uri = smart_open_webhdfs.convert_to_http_uri(parsed_uri)
+            return smart_open_webhdfs.open(http_uri, mode, **kw), filename
         elif parsed_uri.scheme.startswith('http'):
             #
             # The URI may contain a query string and fragments, which interfere
@@ -572,6 +583,9 @@ def _open_binary_stream(uri, mode, transport_params):
             filename = P.basename(urlparse.urlparse(uri).path)
             kw = _check_kwargs(smart_open_http.open, transport_params)
             return smart_open_http.open(uri, mode, **kw), filename
+        elif parsed_uri.scheme == smart_open_gcs.SUPPORTED_SCHEME:
+            kw = _check_kwargs(smart_open_gcs.open, transport_params)
+            return smart_open_gcs.open(parsed_uri.bucket_id, parsed_uri.blob_id, mode, **kw), filename
         else:
             raise NotImplementedError("scheme %r is not supported", parsed_uri.scheme)
     elif hasattr(uri, 'read'):
@@ -666,12 +680,12 @@ def _my_urlsplit(url):
     https://github.com/python/cpython/blob/3.7/Lib/urllib/parse.py
     https://github.com/RaRe-Technologies/smart_open/issues/285
     """
-    if '?' not in url:
-        return urlsplit(url, allow_fragments=False)
+    parsed_url = urlparse.urlsplit(url, allow_fragments=False)
+    if parsed_url.scheme not in smart_open_s3.SUPPORTED_SCHEMES or '?' not in url:
+        return parsed_url
 
-    sr = urlsplit(url.replace('?', '\n'), allow_fragments=False)
-    SplitResult = collections.namedtuple('SplitResult', 'scheme netloc path query fragment')
-    return SplitResult(sr.scheme, sr.netloc, sr.path.replace('\n', '?'), '', '')
+    sr = urlparse.urlsplit(url.replace('?', '\n'), allow_fragments=False)
+    return urlparse.SplitResult(sr.scheme, sr.netloc, sr.path.replace('\n', '?'), '', '')
 
 
 def _parse_uri(uri_as_string):
@@ -681,6 +695,7 @@ def _parse_uri(uri_as_string):
     Supported URI schemes are:
 
       * file
+      * gs
       * hdfs
       * http
       * https
@@ -708,6 +723,7 @@ def _parse_uri(uri_as_string):
       * file:///home/user/file.bz2
       * [ssh|scp|sftp]://username@host//path/file
       * [ssh|scp|sftp]://username@host/path/file
+      * gs://my_bucket/my_blob
 
     """
     if os.name == 'nt':
@@ -721,7 +737,7 @@ def _parse_uri(uri_as_string):
     if parsed_uri.scheme == "hdfs":
         return _parse_uri_hdfs(parsed_uri)
     elif parsed_uri.scheme == "webhdfs":
-        return _parse_uri_webhdfs(parsed_uri)
+        return parsed_uri
     elif parsed_uri.scheme in smart_open_s3.SUPPORTED_SCHEMES:
         return _parse_uri_s3x(parsed_uri)
     elif parsed_uri.scheme == 'file':
@@ -732,6 +748,8 @@ def _parse_uri(uri_as_string):
         return Uri(scheme=parsed_uri.scheme, uri_path=uri_as_string)
     elif parsed_uri.scheme in smart_open_ssh.SCHEMES:
         return _parse_uri_ssh(parsed_uri)
+    elif parsed_uri.scheme == smart_open_gcs.SUPPORTED_SCHEME:
+        return _parse_uri_gcs(parsed_uri)
     else:
         raise NotImplementedError(
             "unknown URI scheme %r in %r" % (parsed_uri.scheme, uri_as_string)
@@ -748,17 +766,6 @@ def _parse_uri_hdfs(parsed_uri):
     return Uri(scheme='hdfs', uri_path=uri_path)
 
 
-def _parse_uri_webhdfs(parsed_uri):
-    assert parsed_uri.scheme == 'webhdfs'
-    uri_path = parsed_uri.netloc + "/webhdfs/v1" + parsed_uri.path
-    if parsed_uri.query:
-        uri_path += "?" + parsed_uri.query
-    if not uri_path:
-        raise RuntimeError("invalid WebHDFS URI: %s" % str(parsed_uri))
-
-    return Uri(scheme='webhdfs', uri_path=uri_path)
-
-
 def _parse_uri_s3x(parsed_uri):
     #
     # Restrictions on bucket names and labels:
@@ -775,7 +782,7 @@ def _parse_uri_s3x(parsed_uri):
     assert parsed_uri.scheme in smart_open_s3.SUPPORTED_SCHEMES
 
     port = 443
-    host = boto.config.get('s3', 'host', _DEFAULT_S3_HOST)
+    host = _DEFAULT_S3_HOST
     ordinary_calling_format = False
     #
     # These defaults tell boto3 to look for credentials elsewhere
@@ -825,24 +832,25 @@ def _parse_uri_file(input_path):
 
 def _parse_uri_ssh(unt):
     """Parse a Uri from a urllib namedtuple."""
-    if '@' in unt.netloc:
-        user, host_port = unt.netloc.split('@', 1)
-    else:
-        user, host_port = None, unt.netloc
+    return Uri(
+        scheme=unt.scheme,
+        uri_path=_unquote(unt.path),
+        user=_unquote(unt.username),
+        host=unt.hostname,
+        port=int(unt.port or smart_open_ssh.DEFAULT_PORT),
+        password=_unquote(unt.password),
+    )
 
-    if ':' in host_port:
-        host, port = host_port.split(':', 1)
-    else:
-        host, port = host_port, None
 
-    if not user:
-        user = None
-    if not port:
-        port = smart_open_ssh.DEFAULT_PORT
-    else:
-        port = int(port)
+def _unquote(text):
+    return text and urlparse.unquote(text)
 
-    return Uri(scheme=unt.scheme, uri_path=unt.path, user=user, host=host, port=port)
+
+def _parse_uri_gcs(parsed_uri):
+    assert parsed_uri.scheme == smart_open_gcs.SUPPORTED_SCHEME
+    bucket_id, blob_id = parsed_uri.netloc, parsed_uri.path[1:]
+
+    return Uri(scheme=parsed_uri.scheme, bucket_id=bucket_id, blob_id=blob_id)
 
 
 def _need_to_buffer(file_obj, mode, ext):
