@@ -23,6 +23,7 @@ import six
 
 import smart_open
 from smart_open import smart_open_lib
+from smart_open import webhdfs
 
 logger = logging.getLogger(__name__)
 
@@ -39,7 +40,7 @@ class ParseUriTest(unittest.TestCase):
     def test_scheme(self):
         """Do URIs schemes parse correctly?"""
         # supported schemes
-        for scheme in ("s3", "s3a", "s3n", "hdfs", "file", "http", "https"):
+        for scheme in ("s3", "s3a", "s3n", "hdfs", "file", "http", "https", "gs"):
             parsed_uri = smart_open_lib._parse_uri(scheme + "://mybucket/mykey")
             self.assertEqual(parsed_uri.scheme, scheme)
 
@@ -136,17 +137,29 @@ class ParseUriTest(unittest.TestCase):
             "s3://access_id@access_secret@mybucket@port/mykey",
         )
 
-    def test_webhdfs_uri(self):
-        """Do webhdfs URIs parse correctly"""
-        # valid uri, no query
-        parsed_uri = smart_open_lib._parse_uri("webhdfs://host:port/path/file")
-        self.assertEqual(parsed_uri.scheme, "webhdfs")
-        self.assertEqual(parsed_uri.uri_path, "host:port/webhdfs/v1/path/file")
+    def test_webhdfs_uri_to_http(self):
+        parsed_uri = smart_open_lib._parse_uri("webhdfs://host:14000/path/file")
+        actual = webhdfs.convert_to_http_uri(parsed_uri)
+        expected = "http://host:14000/webhdfs/v1/path/file"
+        self.assertEqual(actual, expected)
 
-        # valid uri, with query
-        parsed_uri = smart_open_lib._parse_uri("webhdfs://host:port/path/file?query_part_1&query_part2")
-        self.assertEqual(parsed_uri.scheme, "webhdfs")
-        self.assertEqual(parsed_uri.uri_path, "host:port/webhdfs/v1/path/file?query_part_1&query_part2")
+    def test_webhdfs_uri_to_http_with_query(self):
+        parsed_uri = smart_open_lib._parse_uri("webhdfs://host:14000/path/file?a=1")
+        actual = webhdfs.convert_to_http_uri(parsed_uri)
+        expected = "http://host:14000/webhdfs/v1/path/file?a=1"
+        self.assertEqual(actual, expected)
+
+    def test_webhdfs_uri_to_http_with_user(self):
+        parsed_uri = smart_open_lib._parse_uri("webhdfs://user@host:14000/path")
+        actual = webhdfs.convert_to_http_uri(parsed_uri)
+        expected = "http://host:14000/webhdfs/v1/path?user.name=user"
+        self.assertEqual(actual, expected)
+
+    def test_webhdfs_uri_to_http_with_user_and_query(self):
+        parsed_uri = smart_open_lib._parse_uri("webhdfs://user@host:14000/path?a=1")
+        actual = webhdfs.convert_to_http_uri(parsed_uri)
+        expected = "http://host:14000/webhdfs/v1/path?a=1&user.name=user"
+        self.assertEqual(actual, expected)
 
     def test_uri_from_issue_223_works(self):
         uri = "s3://:@omax-mis/twilio-messages-media/final/MEcd7c36e75f87dc6dd9e33702cdcd8fb6"
@@ -203,6 +216,17 @@ class ParseUriTest(unittest.TestCase):
         self.assertEqual(uri.user, 'user')
         self.assertEqual(uri.host, 'host')
         self.assertEqual(uri.port, 1234)
+        self.assertEqual(uri.password, None)
+
+    def test_ssh_with_pass(self):
+        as_string = 'ssh://user:pass@host:1234/path/to/file'
+        uri = smart_open_lib._parse_uri(as_string)
+        self.assertEqual(uri.scheme, 'ssh')
+        self.assertEqual(uri.uri_path, '/path/to/file')
+        self.assertEqual(uri.user, 'user')
+        self.assertEqual(uri.host, 'host')
+        self.assertEqual(uri.port, 1234)
+        self.assertEqual(uri.password, 'pass')
 
     def test_scp(self):
         as_string = 'scp://user@host:/path/to/file'
@@ -212,6 +236,17 @@ class ParseUriTest(unittest.TestCase):
         self.assertEqual(uri.user, 'user')
         self.assertEqual(uri.host, 'host')
         self.assertEqual(uri.port, 22)
+        self.assertEqual(uri.password, None)
+
+    def test_scp_with_pass(self):
+        as_string = 'scp://user:pass@host:/path/to/file'
+        uri = smart_open_lib._parse_uri(as_string)
+        self.assertEqual(uri.scheme, 'scp')
+        self.assertEqual(uri.uri_path, '/path/to/file')
+        self.assertEqual(uri.user, 'user')
+        self.assertEqual(uri.host, 'host')
+        self.assertEqual(uri.port, 22)
+        self.assertEqual(uri.password, 'pass')
 
     def test_sftp(self):
         as_string = 'sftp://host/path/to/file'
@@ -221,6 +256,36 @@ class ParseUriTest(unittest.TestCase):
         self.assertEqual(uri.user, None)
         self.assertEqual(uri.host, 'host')
         self.assertEqual(uri.port, 22)
+        self.assertEqual(uri.password, None)
+
+    def test_sftp_with_user_and_pass(self):
+        as_string = 'sftp://user:pass@host:2222/path/to/file'
+        uri = smart_open_lib._parse_uri(as_string)
+        self.assertEqual(uri.scheme, 'sftp')
+        self.assertEqual(uri.uri_path, '/path/to/file')
+        self.assertEqual(uri.user, 'user')
+        self.assertEqual(uri.host, 'host')
+        self.assertEqual(uri.port, 2222)
+        self.assertEqual(uri.password, 'pass')
+
+    def test_ssh_complex_password_with_colon(self):
+        as_string = 'sftp://user:some:complex@password$$@host:2222/path/to/file'
+        uri = smart_open_lib._parse_uri(as_string)
+        self.assertEqual(uri.password, 'some:complex@password$$')
+
+    def test_gs_uri(self):
+        """Do GCS URIs parse correctly?"""
+        # correct uri without credentials
+        parsed_uri = smart_open_lib._parse_uri("gs://mybucket/myblob")
+        self.assertEqual(parsed_uri.scheme, "gs")
+        self.assertEqual(parsed_uri.bucket_id, "mybucket")
+        self.assertEqual(parsed_uri.blob_id, "myblob")
+
+    def test_gs_uri_contains_slash(self):
+        parsed_uri = smart_open_lib._parse_uri("gs://mybucket/mydir/myblob")
+        self.assertEqual(parsed_uri.scheme, "gs")
+        self.assertEqual(parsed_uri.bucket_id, "mybucket")
+        self.assertEqual(parsed_uri.blob_id, "mydir/myblob")
 
 
 class SmartOpenHttpTest(unittest.TestCase):
@@ -231,14 +296,19 @@ class SmartOpenHttpTest(unittest.TestCase):
     @mock.patch('smart_open.ssh.open')
     def test_read_ssh(self, mock_open):
         """Is SSH line iterator called correctly?"""
-        obj = smart_open.smart_open("ssh://ubuntu@ip_address:1022/some/path/lines.txt")
+        obj = smart_open.smart_open(
+            "ssh://ubuntu:pass@ip_address:1022/some/path/lines.txt",
+            hello='world',
+        )
         obj.__iter__()
         mock_open.assert_called_with(
             '/some/path/lines.txt',
             'rb',
             host='ip_address',
             user='ubuntu',
-            port=1022
+            password='pass',
+            port=1022,
+            transport_params={'hello': 'world'},
         )
 
     @responses.activate
@@ -1424,7 +1494,7 @@ class S3OpenTest(unittest.TestCase):
             actual = fin.read()
         self.assertEqual(text, actual)
 
-    @mock.patch('smart_open.s3.SeekableBufferedInputBase')
+    @mock.patch('smart_open.s3.Reader')
     def test_transport_params_is_not_mutable(self, mock_open):
         smart_open.open('s3://access_key:secret_key@host@bucket/key')
         smart_open.open('s3://bucket/key')
@@ -1437,7 +1507,7 @@ class S3OpenTest(unittest.TestCase):
         self.assertIsNone(mock_open.call_args_list[1][1]['session'])
         self.assertIsNotNone(mock_open.call_args_list[0][1]['session'])
 
-    @mock.patch('smart_open.s3.SeekableBufferedInputBase')
+    @mock.patch('smart_open.s3.Reader')
     def test_respects_endpoint_url_read(self, mock_open):
         url = 's3://key_id:secret_key@play.min.io:9000@smart-open-test/README.rst'
         smart_open.open(url)
@@ -1445,7 +1515,7 @@ class S3OpenTest(unittest.TestCase):
         expected = {'endpoint_url': 'https://play.min.io:9000'}
         self.assertEqual(mock_open.call_args[1]['resource_kwargs'], expected)
 
-    @mock.patch('smart_open.s3.BufferedOutputBase')
+    @mock.patch('smart_open.s3.MultipartWriter')
     def test_respects_endpoint_url_write(self, mock_open):
         url = 's3://key_id:secret_key@play.min.io:9000@smart-open-test/README.rst'
         smart_open.open(url, 'wb')
